#case
#include <16F819.h>

#define __encoder__
#device adc=10   
#FUSES INTRC_IO   
#FUSES WDT         
#FUSES NOPROTECT  
#FUSES NOBROWNOUT 
#FUSES MCLR 
#FUSES NOCPD  
#FUSES NOPUT  
#FUSES NOLVP

#use delay(clock=4000000, RESTART_WDT)
#use i2c(Slave,Fast,sda=PIN_B1,scl=PIN_B4,restart_wdt,force_hw,address=0xAB)

const char init1[] = "HEAT BED CONTROL";
#ifdef __encoder__
const char init2[] = "     v3.5e      ";
#else
const char init2[] = "     v3.5       ";
#endif

const char gen1[]  = "Target:   00.00C";
const char gen2[]  = "Current:  00.00C";

volatile unsigned int16 i;
volatile unsigned int8 j;
volatile unsigned int16 adc_raw;
volatile unsigned int16 adc_avg;
volatile unsigned int16 target;
volatile unsigned int16 current;

volatile unsigned int16 DC_OFF;
volatile unsigned int16 DC_ON;
volatile unsigned int1  DC_OUT;


//204-GT thermistor with 10K in voltage divider
#define TEMPS 28
const unsigned int16 thermTable[TEMPS] = {
1010, 1000, 985, 963, 932, 889, 834, 768, 692, 610, //0-90
526, 446, 373, 308, 252, 206, 168, 137, 112, 91, //100-190
75, 62, 51, 43, 36, 30, 26, 22 //  200-270
};


BYTE incoming, state;
BYTE address;
BYTE buffer[4]; //[0] + [1]<<8 = TARGET TEMP; [2] + [3]<<8 = CURRENT TEMP
BYTE changed;

#int_SSP
void  SSP_isr(void) {
  state = i2c_isr_state();
  if(state == 1){  
    incoming = i2c_read();
    address = incoming;
    if (address > 3) address = 0;
  }

  if(state == 2){  
    incoming = i2c_read();
    if (address > 1) address = 0; //only 0 and 1 are writable
    buffer[address] = incoming;
    if (address == 1) changed = 1;
  }

  if(state == 0x80){
    i2c_write (buffer[address]);
  }
}


#INT_TIMER1 
void timer1_routine(){
  disable_interrupts(INT_TIMER1);
  if(DC_OUT == 1){ 
    DC_OUT = 0;
    output_low(PIN_B2);
    set_timer1(DC_OFF);
  } else {
    DC_OUT = 1;
    output_high(PIN_B2);
    set_timer1(DC_ON);
  }
  enable_interrupts(INT_TIMER1);
}

#ifdef __encoder__
volatile int8 encoder;

#int_ext
void detect_rb0_change() { //here we catch only RB0
  encoder = 1;
  if (input(PIN_B0)) { 
    if (!input(PIN_B5)) {  
      //10
      if(target<27000) target -= 50;
    } else {
      //11
      if(target>49) target += 50;
    }
  } else  { 
    if (input(PIN_B5)) {   
      //01
      if(target<27000) target -= 50;
    } 
    else {
      //00
      if(target>49) target += 50;
    }
  }
}

#endif

//RB3 - RS
struct lcd_pin_map {                 
  BOOLEAN unused;       
  int     data : 4;   
  BOOLEAN mclr;
  BOOLEAN rw;             
  BOOLEAN enable;          
} lcd;

#locate lcd = getenv("sfr:PORTA") 
#define set_tris_lcd(x) set_tris_a(x)

#define lcd_type 2
#define lcd_line_two 0x40

BYTE const LCD_INIT_STRING[4] = {0x20 | (lcd_type << 2), 0xc, 1, 6};

struct lcd_pin_map const LCD_WRITE = {1, 0,0,0,0}; // For write mode all pins are out
struct lcd_pin_map const LCD_READ  = {1,15,0,0,0}; // For read mode data pins are in

BYTE lcd_read_byte() {
      BYTE low,high;
      set_tris_lcd(LCD_READ);
      lcd.rw = 1;
      delay_cycles(1);
      lcd.enable = 1;
      delay_cycles(1);
      high = lcd.data;
      lcd.enable = 0;
      delay_cycles(1);
      lcd.enable = 1;
      delay_us(1);
      low = lcd.data;
      lcd.enable = 0;
      set_tris_lcd(LCD_WRITE);
      return( (high<<4) | low);
}

void lcd_send_nibble( BYTE n ) {
      lcd.data = n;
      delay_cycles(1);
      lcd.enable = 1;
      delay_us(2);
      lcd.enable = 0;
}

void lcd_send_byte( BYTE address, BYTE n ) {
      //lcd.rs = 0;
      output_low(PIN_B3);
      while ( bit_test(lcd_read_byte(),7) ) ;
      //lcd.rs = address;
      if (address){
        output_high(PIN_B3);
      }else{
        output_low(PIN_B3);
      }
      delay_cycles(1);
      lcd.rw = 0;
      delay_cycles(1);
      lcd.enable = 0;
      lcd_send_nibble(n >> 4);
      lcd_send_nibble(n & 0xf);
}

void lcd_init() {
    BYTE i;
    set_tris_lcd(LCD_WRITE);
    //lcd.rs = 0;
    output_low(PIN_B3);
    lcd.rw = 0;
    lcd.enable = 0;
    delay_ms(15);
    for(i=1;i<=3;++i) {
       lcd_send_nibble(3);
       delay_ms(5);
    }
    lcd_send_nibble(2);
    for(i=0;i<=3;++i)
       lcd_send_byte(0,LCD_INIT_STRING[i]);
}

unsigned int16 getTemp(){
  // adc_avg -raw data from adc 10bit
  // thermTable -data from 0 - 270deg every 10 deg
  unsigned int16 tempVal;
  unsigned int16 lowValue;
  unsigned int16 highValue;
  unsigned int16 dataValue;
  unsigned int16 range;

  for(j=1; j<TEMPS; ++j){
    tempVal = thermTable[j];
    if (adc_avg > tempVal){
      restart_wdt();
      lowValue = tempVal;
      highValue = thermTable[j-1];
      range = highValue - lowValue;
      dataValue = adc_avg - lowValue;
      return j * 1000 - (dataValue * 1000 / range);
    }
  }
  return (55555);
}

void print_temp(int8 line, unsigned int16 temp){  // temperature is *100
  unsigned int8 x;

  //hundreds
  x = temp / 10000;
  if (x == 0) {
    lcd_send_byte(0,0x80|(9+lcd_line_two*line)); 
    lcd_send_byte(1,' ');
  } else {
    x = x + 48;
    lcd_send_byte(0,0x80|(9+lcd_line_two*line)); 
    lcd_send_byte(1, x);
  }
  restart_wdt();

  
  //tens
  x = ((temp / 1000 ) % 10) + 48;
  lcd_send_byte(0,0x80|(10+lcd_line_two*line)); 
  lcd_send_byte(1, x);
  restart_wdt();

  //ones
  x = ((temp / 100 ) % 10) + 48;
  lcd_send_byte(0,0x80|(11+lcd_line_two*line)); 
  lcd_send_byte(1, x);
  restart_wdt();

  // 1/10
  x = ((temp / 10 ) % 10) + 48;
  lcd_send_byte(0,0x80|(13+lcd_line_two*line)); 
  lcd_send_byte(1, x);
  restart_wdt();

  // 1/100
  x = (temp  % 10) + 48;
  lcd_send_byte(0,0x80|(14+lcd_line_two*line)); 
  lcd_send_byte(1, x);
  restart_wdt();

}

void h_off(){
  DC_ON  = 0x0000; 
  DC_OFF = 0xFFFD; 
  DC_OUT = 1;
  output_high(PIN_B2);
  set_timer1(DC_ON); 
}

void h_on(){
  DC_ON  = 0xFFFD; 
  DC_OFF = 0x0000; 
  DC_OUT = 0;
  output_low(PIN_B2);
  set_timer1(DC_OFF);
}

void h_75(){ 
 DC_ON  = 0xB6C1; //150ms
 DC_OFF = 0x2445; //450ms
 DC_OUT = 0; 
 output_low(PIN_B2);
 set_timer1(DC_OFF); 
}

void h_50(){ 
 DC_ON  = 0x6D83; //300ms
 DC_OFF = 0x6D83; //300ms
 DC_OUT = 0; 
 output_low(PIN_B2);
 set_timer1(DC_OFF); 
}

void h_25(){ 
 DC_ON  = 0x2445; //450ms
 DC_OFF = 0xB6C1; //150ms
 DC_OUT = 0; 
 output_low(PIN_B2);
 set_timer1(DC_OFF); 
}

void h_5(){ 
 DC_ON  = 0x0000; //528ms
 DC_OFF = 0xF291; //27.5ms
 DC_OUT = 0; 
 output_high(PIN_B2);
 set_timer1(DC_OFF); 
}

void pid(){ 
  static int8 DC = 0;
  
  if (current >= target) {
    h_off();    
    DC = 0;
  } else if (current > (target -  700)) {
    if (DC != 5){
      h_5();
      DC = 5;
    }
  } else if (current > (target - 1300)) {
    if (DC != 25){
      h_25();
      DC = 25;
    }
  } else if (current > (target - 2000)) {
    if (DC != 50){
      h_50();
      DC = 50;
    }
  } else if (current > (target - 2500)) {
    if (DC != 75){
      h_75();
      DC = 75;
    }
  } else  {
    h_on();
    DC = 100;
  }  
}


#ifndef __encoder__
//RB0, RB5
void menu(){
  delay_ms(15);
  if (!input(PIN_B0)) if(target<27000) target += 50;
  if (!input(PIN_B5)) if(target>49) target -= 50;
  
  if (!input(PIN_B0) && !input(PIN_B5)){
    write_eeprom(0, target & 0x00ff);
    write_eeprom(1, target >> 8 );
  }
  print_temp(0, target);
  delay_ms(15);
}
#endif

void main(){
   
   setup_oscillator(OSC_4MHZ); 
   setup_adc_ports(AN0);
   //setup_adc(ADC_CLOCK_INTERNAL);
   setup_adc(ADC_CLOCK_DIV_16);


   //setup_timer_0(RTCC_INTERNAL|RTCC_DIV_1);
   setup_wdt(WDT_2304MS);
   setup_timer_1(T1_INTERNAL | T1_DIV_BY_8); 
   //FOSC/4/8 = 125KHz (1sec = 125,000 ticks, 1tick = 8usec)
   // 0x0000 = 0.52428 sec
   // 0xFFFF = 0sec
   // time = (0xFFFF-X)*8 (usec) // X = 0xFFFF - time(usec)/8
   // 1ms = 125 ticks
   // 
   // 5%  (555ms)
   // 27.5:528 (ms) -- F291:0
   //
   // 25% (600ms)
   // 150:450  (ms) -- B6C1:2445
   //
   // 50% (600ms)
   // 300:300  (ms) -- 6D83:6D83
   //
   // 75% (600ms)
   // 450:150  (ms) -- 2445:B6C1
   
   
   DC_OFF = 0xFFF0;
   DC_ON  = 0x0000;
   DC_OUT = 1; 
   output_high(PIN_B2);
   set_timer1(DC_ON);
   

   set_tris_b(0x33);
   lcd_init();
   output_high(PIN_B2);

   for (j=0;j<16;j++){
     lcd_send_byte(0,0x80|j); 
     lcd_send_byte(1,init1[j]); 

     lcd_send_byte(0,0x80|(j+lcd_line_two)); 
     lcd_send_byte(1,init2[j]); 
   }
   restart_wdt();
   changed = 0;
   address = 0;
   target = read_eeprom(1);
   target = target << 8;
   target |= read_eeprom(0);
   if (target > 30000) target = 7000; //70C
   current = 0;
   adc_raw = 0;
   adc_avg = 0;
   buffer[0]=buffer[1]=buffer[2]=buffer[3]=0;
   delay_ms(1000);

   for (j=0;j<16;j++){
     lcd_send_byte(0,0x80|j);
     lcd_send_byte(1,gen1[j]); 

     lcd_send_byte(0,0x80|(j+lcd_line_two)); 
     lcd_send_byte(1,gen2[j]); 
   }

   print_temp(0, target);
   set_adc_channel(0);
#ifdef __encoder__
   encoder = 0;
   enable_interrupts(INT_EXT);//RB0
#endif
   enable_interrupts(INT_SSP);
   enable_interrupts(INT_TIMER1);
   enable_interrupts(GLOBAL);
   
   while (TRUE){
     restart_wdt();

     adc_raw = read_adc();
     adc_avg = (adc_avg  + adc_raw*3) /4 ;
     current = getTemp(); 

     pid();                   
     print_temp(1, current);  
#ifdef __encoder__
     if(encoder){
        print_temp(0, target);
        encoder = 0;
     }
#else
     if (!input(PIN_B0) || !input(PIN_B5)) menu();
#endif
     if (changed){ //something came via I2C
        target = buffer[1];
        target = target << 8;
        target |= buffer[0];
        print_temp(0, target);
        changed = 0;        
     } 
     //write data into buffer for I2C read
     buffer[2] = current & 0x00ff;
     buffer[3] = current >> 8;
   }
}

